
from PySide6 import QtGui, QtWidgets, QtCore, QtMultimedia, QtMultimediaWidgets
from PySide6.QtOpenGLWidgets import QOpenGLWidget as QOpenGLWidget
from PySide6.QtCore import Signal, Slot

QLabel = QtWidgets.QLabel
QApplication = QtWidgets.QApplication
QTimer = QtCore.QTimer
